/* -*-c-*-
 *
 * Handling of symmetric keysets
 *
 * (c) 2001 Straylight/Edgeware
 */

/*----- Licensing notice --------------------------------------------------*
 *
 * This file is part of Trivial IP Encryption (TrIPE).
 *
 * TrIPE is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * TrIPE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with TrIPE; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*----- Header files ------------------------------------------------------*/

#include "tripe.h"

/*----- Handy macros ------------------------------------------------------*/

#define KEYOK(ks, now) ((ks)->sz_exp > 0 && (ks)->t_exp > now)

/*----- Low-level packet encryption and decryption ------------------------*/

/* --- Encrypted data format --- *
 *
 * Let %$p_i$% be the %$i$%-th plaintext message, with type %$t$%.  We first
 * compute
 *
 *   %$c_i = \mathcal{E}\textrm{-CBC}_{K_{\text{E}}}(p_i)$%
 *
 * as the CBC-ciphertext of %$p_i$%, and then
 *
 *   %$\sigma_i = \mathcal{T}_{K_{\text{M}}}(t, i, c_i)$%
 *
 * as a MAC on the %%\emph{ciphertext}%%.  The message sent is then the pair
 * %$(\sigma_i, c_i)$%.  This construction is provably secure in the NM-CCA
 * sense (assuming that the cipher is IND-CPA, and the MAC is SUF-CMA)
 * [Bellare and Namprempre].
 *
 * This also ensures that, assuming the key is good, we have a secure channel
 * [Krawczyk].  Actually, [Krawczyk] shows that, if the cipher is either a
 * simple stream cipher or a block cipher in CBC mode, we can use the MAC-
 * then-encrypt scheme and still have a secure channel.  However, I like the
 * NM-CCA guarantee from [Bellare and Namprempre].  I'm less worried about
 * the Horton Principle [Wagner and Schneier].
 */

/* --- @doencrypt@ --- *
 *
 * Arguments:	@keyset *ks@ = pointer to keyset to use
 *		@unsigned ty@ = type of message this is
 *		@buf *b@ = pointer to an input buffer
 *		@buf *bb@ = pointer to an output buffer
 *
 * Returns:	Zero if OK; @KSERR_REGEN@ if it's time to generate new keys.
 *		Also returns zero if there was insufficient buffer space, but
 *		the buffer is broken in this case.
 *
 * Use:		Encrypts a message with the given key.  We assume that the
 *		keyset is OK to use.
 */

static int doencrypt(keyset *ks, unsigned ty, buf *b, buf *bb)
{
  int rc;
  size_t sz = BLEFT(b);
  size_t osz, nsz;

  /* --- Initial tracing --- */

  IF_TRACING(T_KEYSET, {
    trace(T_KEYSET,
	  "keyset: encrypting packet %lu (type %u) using keyset %u",
	  (unsigned long)ks->oseq, ty, ks->seq);
    trace_block(T_CRYPTO, "crypto: plaintext packet", BCUR(b), sz);
  })

  /* --- Apply the bulk-crypto transformation --- */

  rc = ks->bulk->encrypt(ks, ty, b, bb);
  if (rc || !BOK(bb)) return (rc);

  /* --- Do the necessary accounting for data volume --- */

  osz = ks->sz_exp;
  nsz = osz > sz ? osz - sz : 0;
  if (osz >= ks->sz_regen && ks->sz_regen > nsz) {
    T( trace(T_KEYSET, "keyset: keyset %u data regen limit exceeded -- "
	     "forcing exchange", ks->seq); )
    rc = KSERR_REGEN;
  }
  ks->sz_exp = nsz;

  /* --- We're done --- */

  return (rc);
}

/* --- @dodecrypt@ --- *
 *
 * Arguments:	@keyset *ks@ = pointer to keyset to use
 *		@unsigned ty@ = expected type code
 *		@buf *b@ = pointer to an input buffer
 *		@buf *bb@ = pointer to an output buffer
 *		@uint32 *seq@ = where to store the sequence number
 *
 * Returns:	Zero on success; @KSERR_DECRYPT@ on failure.
 *
 * Use:		Attempts to decrypt a message with the given key.  No other
 *		checking (e.g., sequence number checks) is performed.  We
 *		assume that the keyset is OK to use, and that there is
 *		sufficient output buffer space reserved.  If the decryption
 *		is successful, the buffer pointer is moved past the decrypted
 *		packet, and the packet's sequence number is stored in @*seq@.
 */

static int dodecrypt(keyset *ks, unsigned ty, buf *b, buf *bb, uint32 *seq)
{
  const octet *q = BCUR(bb);
  int rc;

  IF_TRACING(T_KEYSET, {
    trace(T_KEYSET,
	  "keyset: try decrypting packet (type %u) using keyset %u",
	  ty, ks->seq);
    trace_block(T_CRYPTO, "crypto: ciphertext packet", BCUR(b), BLEFT(b));
  })

  rc = ks->bulk->decrypt(ks, ty, b, bb, seq);
  if (rc) return (rc);

  IF_TRACING(T_KEYSET, {
    trace(T_KEYSET, "keyset: decrypted OK (sequence = %lu)",
	  (unsigned long)*seq);
    trace_block(T_CRYPTO, "crypto: decrypted packet", q, BCUR(bb) - q);
  })
  return (0);
}

/*----- Operations on a single keyset -------------------------------------*/

/* --- @ks_drop@ --- *
 *
 * Arguments:	@keyset *ks@ = pointer to a keyset
 *
 * Returns:	---
 *
 * Use:		Decrements a keyset's reference counter.  If the counter hits
 *		zero, the keyset is freed.
 */

void ks_drop(keyset *ks)
{
  if (--ks->ref)
    return;

#define DROP(dir, a, drop) do { if (ks->dir.a) drop(ks->dir.a); } while (0)
#define DROP_DIR(dir) do {						\
  DROP(dir, c, GC_DESTROY);						\
  DROP(dir, m, GM_DESTROY);						\
} while (0)

  DROP_DIR(in);
  DROP_DIR(out);

#undef DROP
#undef DROP_DIR

  DESTROY(ks);
}

/* --- @ks_gen@ --- *
 *
 * Arguments:	@const void *k@ = pointer to key material
 *		@size_t x, y, z@ = offsets into key material (see below)
 *		@peer *p@ = pointer to peer information
 *
 * Returns:	A pointer to the new keyset.
 *
 * Use:		Derives a new keyset from the given key material.  The
 *		offsets @x@, @y@ and @z@ separate the key material into three
 *		parts.  Between the @k@ and @k + x@ is `my' contribution to
 *		the key material; between @k + x@ and @k + y@ is `your'
 *		contribution; and between @k + y@ and @k + z@ is a shared
 *		value we made together.  These are used to construct two
 *		pairs of symmetric keys.  Each pair consists of an encryption
 *		key and a message authentication key.  One pair is used for
 *		outgoing messages, the other for incoming messages.
 *
 *		The new key is marked so that it won't be selected for output
 *		by @ksl_encrypt@.  You can still encrypt data with it by
 *		calling @ks_encrypt@ directly.
 */

static void gen_dir(const algswitch *algs, struct ksdir *ksd,
		    const char *whichdir,
		    const octet *from, size_t fromsz,
		    const octet *to, size_t tosz,
		    const octet *both, size_t bothsz)
{
#define SETKEY(what, a, init) do {					\
  ghash *_h;								\
  octet *_hh;								\
									\
  if (!algs->a)								\
    ksd->a = 0;								\
  else {								\
    _h = GH_INIT(algs->h);						\
    HASH_STRING(_h, "tripe-" what);					\
    GH_HASH(_h, from, fromsz);						\
    GH_HASH(_h, to, tosz);						\
    GH_HASH(_h, both, bothsz);						\
    _hh = GH_DONE(_h, 0);						\
    IF_TRACING(T_KEYSET, { IF_TRACING(T_CRYPTO, {			\
      char _buf[32];							\
      sprintf(_buf, "crypto: %s key " what, whichdir);			\
      trace_block(T_CRYPTO, _buf, _hh, algs->a##ksz);			\
    }) })								\
    ksd->a = init(algs->a, _hh, algs->a##ksz);				\
    GH_DESTROY(_h);							\
  }									\
} while (0)

  SETKEY("encryption", c, GC_INIT);
  SETKEY("integrity", m, GM_KEY);
  SETKEY("blkc", b, GC_INIT);

#undef SETKEY
}

keyset *ks_gen(const void *k, size_t x, size_t y, size_t z, peer *p)
{
  keyset *ks = CREATE(keyset);
  time_t now = time(0);
  const octet *pp = k;
  const algswitch *algs = &p->kx.kpriv->algs;
  T( static unsigned seq = 0; )

  T( trace(T_KEYSET, "keyset: adding new keyset %u", seq); )

  gen_dir(algs, &ks->in, "incoming", pp, x, pp + x, y - x, pp + y, z - y);
  gen_dir(algs, &ks->out, "outgoing", pp + x, y - x, pp, x, pp + y, z - y);

  T( ks->seq = seq++; )
  ks->bulk = algs->bulk;
  ks->ref = 1;
  ks->t_exp = now + T_EXP;
  ks->sz_exp = algs->expsz;
  ks->sz_regen = algs->expsz/2;
  ks->oseq = 0;
  seq_reset(&ks->iseq);
  ks->next = 0;
  ks->p = p;
  ks->f = KSF_LISTEN;
  ks->tagsz = algs->tagsz;
  return (ks);
}

/* --- @ks_activate@ --- *
 *
 * Arguments:	@keyset *ks@ = pointer to a keyset
 *
 * Returns:	---
 *
 * Use:		Activates a keyset, so that it can be used for encrypting
 *		outgoing messages.
 */

void ks_activate(keyset *ks)
{
  if (ks->f & KSF_LISTEN) {
    T( trace(T_KEYSET, "keyset: activating keyset %u", ks->seq); )
    ks->f &= ~KSF_LISTEN;
  }
}

/* --- @ks_encrypt@ --- *
 *
 * Arguments:	@keyset *ks@ = pointer to a keyset
 *		@unsigned ty@ = message type
 *		@buf *b@ = pointer to input buffer
 *		@buf *bb@ = pointer to output buffer
 *
 * Returns:	Zero if successful; @KSERR_REGEN@ if we should negotiate a
 *		new key; @KSERR_NOKEYS@ if the key is not usable.  Also
 *		returns zero if there was insufficient buffer (but the output
 *		buffer is broken in this case).
 *
 * Use:		Encrypts a block of data using the key.  Note that the `key
 *		ought to be replaced' notification is only ever given once
 *		for each key.  Also note that this call forces a keyset to be
 *		used even if it's marked as not for data output.
 *
 *		The encryption transform is permitted to corrupt @buf_u@ for
 *		its own purposes.  Neither the source nor destination should
 *		be within @buf_u@; and callers mustn't expect anything stored
 *		in @buf_u@ to still
 */

int ks_encrypt(keyset *ks, unsigned ty, buf *b, buf *bb)
{
  time_t now = time(0);

  if (!KEYOK(ks, now)) {
    buf_break(bb);
    return (KSERR_NOKEYS);
  }
  return (doencrypt(ks, ty, b, bb));
}

/* --- @ks_decrypt@ --- *
 *
 * Arguments:	@keyset *ks@ = pointer to a keyset
 *		@unsigned ty@ = expected type code
 *		@buf *b@ = pointer to an input buffer
 *		@buf *bb@ = pointer to an output buffer
 *
 * Returns:	Zero on success; @KSERR_...@ on failure.  Also returns
 *		zero if there was insufficient buffer (but the output buffer
 *		is broken in this case).
 *
 * Use:		Attempts to decrypt a message using a given key.  Note that
 *		requesting decryption with a key directly won't clear a
 *		marking that it's not for encryption.
 *
 *		The decryption transform is permitted to corrupt @buf_u@ for
 *		its own purposes.  Neither the source nor destination should
 *		be within @buf_u@; and callers mustn't expect anything stored
 *		in @buf_u@ to still
 */

int ks_decrypt(keyset *ks, unsigned ty, buf *b, buf *bb)
{
  time_t now = time(0);
  uint32 seq;
  int err;

  if (!KEYOK(ks, now)) return (KSERR_DECRYPT);
  if (buf_ensure(bb, BLEN(b))) return (0);
  if ((err = dodecrypt(ks, ty, b, bb, &seq)) != 0) return (err);
  if (seq_check(&ks->iseq, seq, "SYMM")) return (KSERR_SEQ);
  return (0);
}

/*----- Keyset list handling ----------------------------------------------*/

/* --- @ksl_free@ --- *
 *
 * Arguments:	@keyset **ksroot@ = pointer to keyset list head
 *
 * Returns:	---
 *
 * Use:		Frees (releases references to) all of the keys in a keyset.
 */

void ksl_free(keyset **ksroot)
{
  keyset *ks, *ksn;
  for (ks = *ksroot; ks; ks = ksn) {
    ksn = ks->next;
    ks->f &= ~KSF_LINK;
    ks_drop(ks);
  }
}

/* --- @ksl_link@ --- *
 *
 * Arguments:	@keyset **ksroot@ = pointer to keyset list head
 *		@keyset *ks@ = pointer to a keyset
 *
 * Returns:	---
 *
 * Use:		Links a keyset into a list.  A keyset can only be on one list
 *		at a time.  Bad things happen otherwise.
 */

void ksl_link(keyset **ksroot, keyset *ks)
{
  assert(!(ks->f & KSF_LINK));
  ks->next = *ksroot;
  *ksroot = ks;
  ks->f |= KSF_LINK;
  ks->ref++;
}

/* --- @ksl_prune@ --- *
 *
 * Arguments:	@keyset **ksroot@ = pointer to keyset list head
 *
 * Returns:	---
 *
 * Use:		Prunes the keyset list by removing keys which mustn't be used
 *		any more.
 */

void ksl_prune(keyset **ksroot)
{
  time_t now = time(0);

  while (*ksroot) {
    keyset *ks = *ksroot;

    if (ks->t_exp <= now) {
      T( trace(T_KEYSET, "keyset: expiring keyset %u (time limit reached)",
	       ks->seq); )
      goto kill;
    } else if (ks->sz_exp == 0) {
      T( trace(T_KEYSET, "keyset: expiring keyset %u (data limit reached)",
	       ks->seq); )
      goto kill;
    } else {
      ksroot = &ks->next;
      continue;
    }

  kill:
    *ksroot = ks->next;
    ks->f &= ~KSF_LINK;
    ks_drop(ks);
  }
}

/* --- @ksl_encrypt@ --- *
 *
 * Arguments:	@keyset **ksroot@ = pointer to keyset list head
 *		@unsigned ty@ = message type
 *		@buf *b@ = pointer to input buffer
 *		@buf *bb@ = pointer to output buffer
 *
 * Returns:	Zero if successful; @KSERR_REGEN@ if it's time to negotiate a
 *		new key; @KSERR_NOKEYS@ if there are no suitable keys
 *		available.  Also returns zero if there was insufficient
 *		buffer space (but the output buffer is broken in this case).
 *
 * Use:		Encrypts a packet.
 */

int ksl_encrypt(keyset **ksroot, unsigned ty, buf *b, buf *bb)
{
  time_t now = time(0);
  keyset *ks = *ksroot;

  for (;;) {
    if (!ks) {
      T( trace(T_KEYSET, "keyset: no suitable keysets found"); )
      buf_break(bb);
      return (KSERR_NOKEYS);
    }
    if (KEYOK(ks, now) && !(ks->f & KSF_LISTEN))
      break;
    ks = ks->next;
  }

  return (doencrypt(ks, ty, b, bb));
}

/* --- @ksl_decrypt@ --- *
 *
 * Arguments:	@keyset **ksroot@ = pointer to keyset list head
 *		@unsigned ty@ = expected type code
 *		@buf *b@ = pointer to input buffer
 *		@buf *bb@ = pointer to output buffer
 *
 * Returns:	Zero on success; @KSERR_DECRYPT@ on failure.  Also returns
 *		zero if there was insufficient buffer (but the output buffer
 *		is broken in this case).
 *
 * Use:		Decrypts a packet.
 */

int ksl_decrypt(keyset **ksroot, unsigned ty, buf *b, buf *bb)
{
  time_t now = time(0);
  keyset *ks;
  uint32 seq;
  int err;

  if (buf_ensure(bb, BLEN(b)))
    return (0);

  for (ks = *ksroot; ks; ks = ks->next) {
    if (!KEYOK(ks, now))
      continue;
    if ((err = dodecrypt(ks, ty, b, bb, &seq)) == 0) {
      if (ks->f & KSF_LISTEN) {
	T( trace(T_KEYSET, "keyset: implicitly activating keyset %u",
		 ks->seq); )
	ks->f &= ~KSF_LISTEN;
      }
      if (seq_check(&ks->iseq, seq, "SYMM"))
	return (KSERR_SEQ);
      else
	return (0);
    }
    if (err != KSERR_DECRYPT) return (err);
  }
  T( trace(T_KEYSET, "keyset: no matching keys, or incorrect MAC"); )
  return (KSERR_DECRYPT);
}

/*----- That's all, folks -------------------------------------------------*/
